from __future__ import absolute_import

from django.apps import AppConfig

from scout_apm.django.instruments.sql import SQLInstrument
from scout_apm.django.instruments.template import TemplateInstrument
from scout_apm.django.config import ConfigAdapter
import scout_apm.core

import logging


logger = logging.getLogger(__name__)


class ScoutApmDjangoConfig(AppConfig):
    name = 'scout_apm'
    verbose_name = 'Scout Apm (Django)'

    def ready(self):
        # Copy django configuration to scout_apm's config
        ConfigAdapter.install()

        # Finish installing the agent. If the agent isn't installed for any
        # reason, return without installing instruments
        installed = scout_apm.core.install()
        if installed is False:
            return

        self.install_middleware()

        # Setup Instruments
        SQLInstrument.install()
        TemplateInstrument.install()

    def install_middleware(self):
        """
        Attempts to insert the ScoutApm middleware as the first middleware
        (first on incoming requests, last on outgoing responses).
        """
        from django.conf import settings

        if isinstance(settings.MIDDLEWARE, tuple):
            settings.MIDDLEWARE = (
                ('scout_apm.django.middleware.MiddlewareTimingMiddleware', ) +
                settings.MIDDLEWARE +
                ('scout_apm.django.middleware.ViewTimingMiddleware', ))
        else:
            settings.MIDDLEWARE.insert(0, 'scout_apm.django.middleware.MiddlewareTimingMiddleware')
            settings.MIDDLEWARE.append('scout_apm.django.middleware.ViewTimingMiddleware')




import scout_apm.api
from scout_apm.core.tracked_request import TrackedRequest


def test_context_manager_instrument():
    # Save TR here, so it doesn't disappear on us when span finishes
    tr = TrackedRequest.instance()

    with scout_apm.api.instrument('Test ContextMgr') as instrument:
        instrument.tag('foo', 'bar')

    span = tr.complete_spans[-1]
    assert(span.operation == 'Custom/Test ContextMgr')


def test_decoration_instrument():
    # Save TR here, so it doesn't disappear on us when span finishes
    tr = TrackedRequest.instance()

    @scout_apm.api.instrument('Test Decorator')
    def test():
        a = 1
    test()

    span = tr.complete_spans[-1]
    assert(span.operation == 'Custom/Test Decorator')


def test_context_manager_instrument_with_kind():
    # Save TR here, so it doesn't disappear on us when span finishes
    tr = TrackedRequest.instance()

    with scout_apm.api.instrument('Get', kind='Redis') as instrument:
        instrument.tag('foo', 'bar')

    span = tr.complete_spans[-1]
    assert(span.operation == 'Redis/Get')


def test_decoration_instrument_with_kind():
    # Save TR here, so it doesn't disappear on us when span finishes
    tr = TrackedRequest.instance()

    @scout_apm.api.instrument('GET example.com', kind='HTTP')
    def test():
        a = 1
    test()

    span = tr.complete_spans[-1]
    assert(span.operation == 'HTTP/GET example.com')


def test_context_manager_default_tags():
    # Save TR here, so it doesn't disappear on us when span finishes
    tr = TrackedRequest.instance()

    with scout_apm.api.instrument('tag test', tags={'x': 99}):
        a = 1

    span = tr.complete_spans[-1]
    assert(len(span.tags) == 1)
    assert(span.tags['x'] == 99)


def test_decoration_default_tags():
    # Save TR here, so it doesn't disappear on us when span finishes
    tr = TrackedRequest.instance()

    @scout_apm.api.instrument('tag test', tags={'x': 99})
    def test():
        a = 1
    test()

    span = tr.complete_spans[-1]
    assert(len(span.tags) == 1)
    assert(span.tags['x'] == 99)



from scout_apm.core.tracked_request import TrackedRequest

from sqlalchemy import event

def instrument_sqlalchemy(engine):
    def before_cursor_execute(conn, cursor, statement, parameters, context, executemany):
        tr = TrackedRequest.instance()
        span = tr.start_span(operation='SQL/Query')
        span.tag('db.statement', statement)

    def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
        tr = TrackedRequest.instance()
        span = tr.current_span()
        if span is not None:
            tr.callset.update(statement, 1, span.duration())
            if tr.callset.should_capture_backtrace(statement) is True:
                span.capture_backtrace()
        tr.stop_span()

    if getattr(engine, "_scout_instrumented", False) != True:
        event.listen(engine, 'before_cursor_execute', before_cursor_execute)
        event.listen(engine, 'after_cursor_execute', after_cursor_execute)
        setattr(engine, "_scout_instrumented", True)



import scout_apm.core
from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.core.config.config import ScoutConfig
from scout_apm.api.context import Context


def includeme(config):
    configs = {}
    pyramid_config = config.get_settings()
    for name in filter(lambda x: x.startswith('SCOUT_'), pyramid_config):
        value = pyramid_config[name]
        clean_name = name.replace('SCOUT_', '').lower()
        configs[clean_name] = value
    ScoutConfig.set(**configs)

    if scout_apm.core.install():
        config.add_tween('scout_apm.pyramid.instruments')


def instruments(handler, registry):
    def scout_tween(request):
        try:
            tr = TrackedRequest.instance()
            span = tr.start_span(operation='Controller/Pyramid')

            # Capture what we can from the request, but never fail
            try:
                Context.add('path', request.path)
                Context.add('user_ip', request.remote_addr)
            except:
                pass

            try:
                response = handler(request)
            except:
                tr.tag('error', 'true')
                raise

            # This happens further down the call chain. So time it starting
            # above, but only name it if it gets to here.
            if request.matched_route is not None:
                tr.mark_real_request()
                span.operation = 'Controller/' + request.matched_route.name

        finally:
            tr.stop_span()

        return response

    return scout_tween




import logging
from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.api.context import Context

# Logging
logger = logging.getLogger(__name__)


class MiddlewareTimingMiddleware:
    """
    Insert as early into the Middleware stack as possible (outermost layers),
    so that other middlewares called after can be timed.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        operation = 'Middleware'

        TrackedRequest.instance().start_span(operation=operation)
        response = self.get_response(request)
        TrackedRequest.instance().stop_span()
        return response

class ViewTimingMiddleware:
    """
    Insert as deep into the middleware stack as possible, ideally wrapping no
    other middleware. Designed to time the View itself
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        """
        Wrap a single incoming request with start and stop calls.
        This will start timing, but relies on the process_view callback to
        capture more details about what view was really called, and other
        similar info.

        If process_view isn't called, then the request will not
        be recorded.  This can happen if a middleware further along the stack
        doesn't call onward, and instead returns a response directly.
        """

        tr = TrackedRequest.instance()
        tr.mark_real_request()

        # This operation name won't be recorded unless changed later in
        # process_view
        operation = 'Unknown'
        tr.start_span(operation=operation)
        response = self.get_response(request)
        tr.stop_span()
        return response

    def process_view(self, request, view_func, view_args, view_kwargs):
        """
        Capture details about the view_func that is about to execute
        """
        try:
            view_name = request.resolver_match._func_path
            span = TrackedRequest.instance().current_span()
            if span is not None:
                span.operation = 'Controller/' + view_name
                Context.add('path', request.path)
                Context.add('user_ip', request.get_host())
                if request.user is not None:
                    Context.add('username', request.user.get_username())
        except:
            pass

    def process_exception(self, request, exception):
        """
        Mark this request as having errored out

        Does not modify or catch or otherwise change the exception thrown
        """
        TrackedRequest.instance().tag('error', 'true')

    #  def process_template_response(self, request, response):
    #      """
    #      """
    #      pass



from scout_apm.core.monkey import monkeypatch_method
from scout_apm.core.tracked_request import TrackedRequest
import scout_apm.sqlalchemy

from sqlalchemy import event
from flask_sqlalchemy import SQLAlchemy

def instrument_sqlalchemy(db):
    @monkeypatch_method(SQLAlchemy)
    def get_engine(original, self, *args, **kwargs):
        engine = original(*args, **kwargs)
        scout_apm.sqlalchemy.instrument_sqlalchemy(engine)
        return engine



import os

from scout_apm.core.git_revision import GitRevision


def test_detects_nothing_returns_string():
    if 'HEROKU_SLUG_COMMIT' in os.environ:
        del(os.environ['HEROKU_SLUG_COMMIT'])
    if 'SCOUT_REVISION_SHA' in os.environ:
        del(os.environ['SCOUT_REVISION_SHA'])
    assert('' == GitRevision().detect())


def test_detect_from_heroku_slug():
    os.environ['HEROKU_SLUG_COMMIT'] = 'FROM_HEROKU_SLUG'
    assert('FROM_HEROKU_SLUG' == GitRevision().detect())


def test_detect_from_scout_env():
    os.environ['SCOUT_REVISION_SHA'] = 'FROM_SCOUT_ENV'
    assert('FROM_SCOUT_ENV' == GitRevision().detect())



from scout_apm.core.tracked_request import TrackedRequest

import sys
import logging


# Logging
logger = logging.getLogger(__name__)


# Python 2 (and very early 3.x) didn't have ContextDecorator, so define it for ourselves
if sys.version_info < (3, 2):
    import functools

    class ContextDecorator(object):
        def __call__(self, f):
            @functools.wraps(f)
            def decorated(*args, **kwds):
                with self:
                    return f(*args, **kwds)
            return decorated
else:
    from contextlib import ContextDecorator


class instrument(ContextDecorator):
    def __init__(self, operation, kind='Custom', tags={}):
        self.operation = kind + '/' + operation
        self.tags = tags

    def __enter__(self):
        tr = TrackedRequest.instance()
        self.span = tr.start_span(operation=self.operation)
        for key, value in self.tags.items():
            self.tag(key, value)
        return self

    def __exit__(self, *exc):
        tr = TrackedRequest.instance()
        tr.stop_span()
        return False

    def tag(self, key, value):
        if self.span is not None:
            self.span.tag(key, value)




from scout_apm.core.tracked_request import TrackedRequest


def test_instance():
    tr1 = TrackedRequest.instance()
    tr2 = TrackedRequest.instance()
    assert(tr1.req_id == tr2.req_id)
    assert(tr1 == tr2)


def test_tag_request():
    tr = TrackedRequest()

    tr.tag("foo", "bar")

    assert(len(tr.tags) == 1)
    assert(tr.tags['foo'] == 'bar')


def test_tag_span():
    tr = TrackedRequest()
    span = tr.start_span()
    span.tag("a", "b")
    tr.stop_span()

    assert(len(tr.complete_spans[0].tags) == 1)
    assert(tr.complete_spans[0].tags['a'] == 'b')


def test_start_span_wires_parents():
    tr = TrackedRequest()
    span1 = tr.start_span()
    span2 = tr.start_span()
    assert(span1.parent is None)
    assert(span2.parent == span1.span_id)



from __future__ import absolute_import

import logging

# Logging
logger = logging.getLogger(__name__)


class NPlusOneCallSet():
    def __init__(self):
        self.items = {}

    def update(self, sql_string, call_count, call_duration):
        self.find_for(sql_string).update(call_count, call_duration)

    def should_capture_backtrace(self, sql_string):
        return self.find_for(sql_string).should_capture_backtrace()

    def find_for(self, sql_string):
        if sql_string not in self.items:
            self.items[sql_string] = NPlusOneCallSetItem(sql_string)
        return self.items[sql_string]


class NPlusOneCallSetItem():
    # Fetch backtraces on this number of same SQL calls
    CALL_COUNT_THRESHOLD = 5

    # Minimum time in seconds before we start performing any work.
    DURATION_THRESHOLD = 150/1000.0

    def __init__(self, sql_string):
        self.sql_string = sql_string
        self.captured = False
        self.call_count = 0
        self.call_duration = 0.0  # In Seconds
        self.past_duration_threshold = False

    def update(self, call_count, call_duration):
        if self.captured is True:
            # No need to do any work if we've already captured a backtrace.
            return
        self.call_count += call_count
        self.call_duration += call_duration

    def is_past_duration_threshold(self):
        if self.past_duration_threshold is True:
            # no need to check again once past
            return True
        self.past_duration_threshold = self.call_duration >= self.__class__.DURATION_THRESHOLD
        return self.past_duration_threshold

    # Call count and call duration must both be past thresholds.
    def should_capture_backtrace(self):
        if (not self.captured) and (self.call_count >= self.__class__.CALL_COUNT_THRESHOLD) and self.is_past_duration_threshold():
            self.captured = True
            return True
        return False



from __future__ import absolute_import

import logging
import os

logger = logging.getLogger(__name__)


# Note that this is called from Config Defaults, which is looked up after the
# configs from environment variables. If you set SCOUT_REVISION_SHA, it will be
# picked up in that step of config, not here.
#
# Similarly, this can be overriden by the code level configuration (from Django
# configs or similar) by using the "revision_sha" key.
class GitRevision():
    def detect(self):
        sha = self.detect_from_env_var() or self.detect_from_heroku()
        return sha or ''

    def detect_from_heroku(self):
        return os.environ.get('HEROKU_SLUG_COMMIT')

    def detect_from_env_var(self):
        return os.environ.get('SCOUT_REVISION_SHA')



from contextlib import ContextDecorator

class instrument(ContextDecorator):
    def __enter__(self):
        print('Starting')
        return self

    def __exit__(self, *exc):
        print('Finishing')
        return False





from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.core.n_plus_one_call_set import NPlusOneCallSet
from scout_apm.core.n_plus_one_call_set import NPlusOneCallSetItem


def test_call_set_finds_or_creates():
    sql = "SELECT * from test"
    callset = NPlusOneCallSet()
    assert(0 == len(callset.items))
    item = callset.find_for(sql)
    assert(1 == len(callset.items))
    assert(item.sql_string == sql)


def test_call_set_updates():
    sql = "SELECT * from test"
    callset = NPlusOneCallSet()
    callset.update(sql, 3, 4)
    item = callset.find_for(sql)
    assert(3 == item.call_count)
    assert(4 == item.call_duration)


def test_should_capture_backtrace():
    sql = "SELECT * from test"
    callset = NPlusOneCallSet()
    callset.update(sql,
                   NPlusOneCallSetItem.CALL_COUNT_THRESHOLD - 1,
                   NPlusOneCallSetItem.DURATION_THRESHOLD - 1)
    assert(callset.should_capture_backtrace(sql) is False)
    callset.update(sql, 1, 1)
    assert(callset.should_capture_backtrace(sql) is True)


def test_no_capture_if_already_captured():
    sql = "SELECT * from test"
    callset = NPlusOneCallSet()
    callset.update(sql,
                   NPlusOneCallSetItem.CALL_COUNT_THRESHOLD,
                   NPlusOneCallSetItem.DURATION_THRESHOLD)
    assert(callset.should_capture_backtrace(sql) is True)
    # Subsequent calls should return False
    assert(callset.should_capture_backtrace(sql) is False)
    callset.update(sql, 1, 1)
    assert(callset.should_capture_backtrace(sql) is False)


def test_is_past_duration_threshold():
    sql = "SELECT * from test"
    callset = NPlusOneCallSet()
    callset.update(sql,
                   0,
                   NPlusOneCallSetItem.DURATION_THRESHOLD - 1)
    item = callset.find_for(sql)
    assert(item.is_past_duration_threshold() is False)
    callset.update(sql, 0, 1)
    assert(item.is_past_duration_threshold() is True)
    # Assert again to test the attribute check in the function
    assert(item.is_past_duration_threshold() is True)



import scout_apm


def test_traceback():
    bt = scout_apm.core.backtrace.capture()
    for frame in bt:
        keys = list(frame.keys())
        keys.sort()
        assert(keys == ['file', 'function', 'line'])



from __future__ import absolute_import

# Python Modules
import logging
from os import getpid

from scout_apm.core.context import AgentContext
from scout_apm.core.core_agent_manager import CoreAgentManager
from scout_apm.core.metadata import AppMetadata

logger = logging.getLogger(__name__)


def install():
    if not AgentContext.instance().config.value('monitor'):
        logger.info("APM Not Launching on PID: %s - Configuration 'monitor' is not true", getpid())
        return False

    logger.debug('APM Launching on PID: %s', getpid())
    CoreAgentManager().launch()
    AppMetadata.report()
    AgentContext.socket().stop()
    return True



import scout_apm.core.commands
from scout_apm.core.tracked_request import TrackedRequest

def test_stop_span_defaults_timestamp_to_invalid_date():
    command = scout_apm.core.commands.StopSpan(request_id="req_id", span_id="span_id")
    assert(command.timestamp == scout_apm.core.commands.INVALID_DATE)


def test_finish_request_defaults_timestamp_to_invalid_date():
    command = scout_apm.core.commands.FinishRequest(request_id="req_id")
    assert(command.timestamp == scout_apm.core.commands.INVALID_DATE)


def test_from_tracked_request_bails_on_unfinished_span():
    tr = TrackedRequest()
    tr.start_span("foo")
    batch = scout_apm.core.commands.BatchCommand.from_tracked_request(tr)
    assert(batch is None)


def test_from_tracked_request_bails_on_unfinished_request():
    tr = TrackedRequest()
    batch = scout_apm.core.commands.BatchCommand.from_tracked_request(tr)
    assert(batch is None)


def test_from_tracked_request_creates_batch_command():
    tr = TrackedRequest()
    tr.start_span("foo")
    tr.stop_span()

    batch = scout_apm.core.commands.BatchCommand.from_tracked_request(tr)
    assert(len(batch.commands) == 4)



from __future__ import absolute_import

from bottle import request

import scout_apm.core
from scout_apm.core.context import AgentContext
from scout_apm.core.config.config import ScoutConfig
from scout_apm.core.tracked_request import TrackedRequest

from scout_apm.api.context import Context


class ScoutPlugin(object):
    def __init__(self):
        self.name = 'scout'
        self.api = 2

    def set_config_from_bottle(self, app):
        scout_config = ScoutConfig()
        bottle_configs = {}
        for k in scout_config.known_keys():
            value = app.config.get("scout.{}".format(k))
            if value is not None and value != '':
                bottle_configs[k] = value
        scout_config.set(**bottle_configs)
        return scout_config

    def setup(self, app):
        self.set_config_from_bottle(app)
        scout_apm.core.install()

    def apply(self, callback, context):
        if AgentContext.instance().config.value('monitor') is not True:
            return callback

        def wrapper(*args, **kwargs):
            try:
                tr = TrackedRequest.instance()
                tr.mark_real_request()
                path = 'Unknown'

                if request.route.name is not None:
                    path = request.route.name
                else:
                    path = request.route.rule

                if path == '/':
                    path = '/home'

                if not path.startswith('/'):
                    path = '/{}'.format(path)

                span = tr.start_span(operation='Controller{}'.format(path))

                try:
                    Context.add('path', path)
                    Context.add('user_ip', request.remote_addr)
                except:
                    pass

                try:
                    response = callback(*args, **kwargs)
                except:
                    tr.tag('error', 'true')
                    raise

            finally:
                tr.stop_span()

            return response
        return wrapper
